#!/usr/bin/python

class BioRangeError(Exception):
    def __init__(self, value):
        self.parameter = value
    def __str__(self):
        return repr(self.parameter)

class BioRange(object):
    '''It represents a range in a Sequence.
    '''
    def __init__(self, start, end, strand=0):
        '''The BioRange constructor. Both begin and end are included in the
        BioRange.

        keyword arguments:
        start -- The fisrt residue of the range (an int)
        end -- The last residue of the range (an int)
        strand -- the strand (default 0)
        '''
        self.set_start(start)
        self.set_end(end)
        self.set_strand(strand)

    def __str__(self):
        toprint = 'start:' + str(self.start) + ' ' 
        toprint += 'end:' + str(self.end) + ' '
        toprint += 'strand:' + str(self.strand) + ' '
        return toprint

    def set_start(self, start):
        '''It sets the start position of the range.

        It should be 0 or positive.
        keyword arguments:
        start -- The fisrt residue of the range
        '''
        if start >= 0:
            self.__start = int(start)
        else:
            raise BioRangeError('The BioRange start position should be an integer equal or greater than 0')

    def get_start(self):
        '''It returns the start position.'''
        return self.__start

    start = property(get_start, set_start)

    def set_end(self, end):
        '''It sets the end position of the range.
    
        It should be equal or greater than the start
        keyword arguments:
        end -- The last residue of the range
        '''
        if end >= self.__start:
            self.__end = end
        else:
            raise BioRangeError('The BioRange end position should be an integer equal or greater than the start')

    def get_end(self):
        '''It returns the end position.'''
        return self.__end

    end = property(get_end, set_end)

    def set_strand(self, strand):
        '''It set the strand.

        It can be -1, 0 or +1.
        '''
        if strand == -1 or strand == 0 or strand == -1:
            self.__strand = strand
        else:
            raise BioRangeError('The BioRange strand should be -1, 0 or +1')

    def get_strand(self):
        '''It returns the strand'''
        return self.__strand

    strand = property(get_strand, set_strand)

    def __len__(self):
        '''It returns the length of the range.'''
        return self.__end - self.__start + 1

    def overlaps(self, range):
        '''It returns True if the ranges overlap, False otherwise.
        
        Keyword argument:
        range -- A BioRange to compare.
        '''
        #self    -----     -----     -----     --------
        #range  ---      ---------    ------    -----
        if self.start >= range.start and self.start <= range.end:
            return True
        if self.end >= range.start and self.end <= range.end:
            return True
        if self.start <= range.start and self.end >= range.end:
            return True
        return False

    def contains(self, range):
        '''It returns True if this range totally contains the given range.
        
        Keyword argument:
        range -- A BioRange to compare.
        '''
        #self  ------------
        #range   --------- 
        if self.start <= range.start and self.end >= range.end:
            return True
        return False

    def equals(self, range):
        '''It returns True if this range is equal to the given range.
        
        Keyword argument:
        range -- A BioRange to compare.
        '''
        #self  ---------
        #range --------- 
        if self.start == range.start and self.end == range.end:
            return True
        return False

    def reverse(self, length=None):
        '''It reverses the range.
        
        To do a proper reverse we need to know the length of the sequence in 
        which this BioRange is. If None is provided we asume that the length
        is equal to the end of the range.

        Keyword arguments:
        length -- the length of the sequence that holds this BioRange,
                  an int (default the end of this BioRange).
        '''
        if length is None:
            length = self.end + 1   #we count from 0
        #external_seq 0123456789
        #range         |---|    
        #offset             |--|
        #new_range        |---| 
        offset = length - self.end - 1
        new_end = length - self.start - 1
        new_start = offset
        self.start = new_start
        self.end = new_end


import unittest

class BioRangeTests(unittest.TestCase):
    def test_basic_functionality(self):
        range = BioRange(0,1)
        assert range.start == 0
        assert range.end == 1
        assert range.strand == 0
        assert len(range) == 2

        range = BioRange(10,20, -1)
        assert  range.start == 10
        assert range.end == 20
        assert range.strand == -1
        assert len(range) == 11

    def test_expections(self):
        self.failUnlessRaises(BioRangeError, BioRange, 1,0)
        self.failUnlessRaises(BioRangeError, BioRange, -1,10)
        self.failUnlessRaises(BioRangeError, BioRange, 10, 20, 2)

    def test_overlaps(self):
        #overlaps
        #   -----
        # ----
        assert BioRange(2,5).overlaps(BioRange(1,3))
        # ----
        #-------
        assert BioRange(4,5).overlaps(BioRange(1,6))
        #-----
        # ------
        assert BioRange(1,5).overlaps(BioRange(2,6))
        #-------
        # ----
        assert BioRange(1,5).overlaps(BioRange(2,4))
        #-------
        #       ------
        assert not BioRange(0,5).overlaps(BioRange(6,10))
        #           -------
        #-------
        assert not BioRange(10,55).overlaps(BioRange(6,8))

        #contains
        #   -----
        # ----
        assert not BioRange(2,5).contains(BioRange(1,3))
        # ----
        #-------
        assert not BioRange(4,5).contains(BioRange(1,6))
        #-----
        # ------
        assert not BioRange(1,5).contains(BioRange(2,6))
        #-------
        # ----
        assert BioRange(1,5).contains(BioRange(2,4))
        #-------
        #       ------
        assert not BioRange(0,5).contains(BioRange(6,10))
        #           -------
        #-------
        assert not BioRange(10,55).contains(BioRange(6,8))

        #equals
        #   -----
        # ----
        assert not BioRange(2,5).equals(BioRange(1,3))
        # ----
        #-------
        assert not BioRange(4,5).equals(BioRange(1,6))
        #-----
        # ------
        assert not BioRange(1,5).equals(BioRange(2,6))
        #-------
        # ----
        assert not BioRange(1,5).equals(BioRange(2,4))
        #-------
        #       ------
        assert not BioRange(0,5).equals(BioRange(6,10))
        #           -------
        #-------
        assert not BioRange(10,55).equals(BioRange(6,8))
        #  --------
        #  --------
        assert BioRange(1,5).equals(BioRange(1,5))

    def test_ismutable(self):
        # 012345
        #  |---|
        # |---|
        range = BioRange(1,5)
        range.reverse()
        assert range.start == 0
        assert range.end == 4

        # 0123456789
        #  |---|
        #     |---|
        range = BioRange(1,5)
        range.reverse(10)
        assert range.start == 4
        assert range.end == 8

        # 01234567890
        #      |---|
        #  |---|
        range = BioRange(5,9)
        range.reverse(11)
        assert range.start == 1
        assert range.end == 5


def main():
    unittest.main()

if __name__ == '__main__':
    main()
