#!/usr/bin/python

from BioRange import BioRange

class RichFeatureError(Exception):
    def __init__(self, value):
        self.parameter = value
    def __str__(self):
        return repr(self.parameter)

class RichFeature(object):
    '''It represents a Feature that can be added to a sequence.

    Several kind of objects can be a feature: str, RichSeq, Seq, list.
    A RichFeature can be associated with a BioRange in the sequence.
    If it is associated with a BioRange and an object with a __len__ method 
    (other than a str) is used, RichFeature will check that the object
    fits into the BioRange.
    '''
    def __init__(self, id=None, type=None, feature=None, range=None):
        '''The RichFeature constructor.

        Keyword arguments:
        id -- The id of the feature (default None)
        type -- The type of the feature (default None)
        feature -- The actual feature (default None)
        range -- The BioRange associated with the feature (default None)
        '''
        self.id = id
        self.type = type
        self.__range = None
        self.set_feature(feature)
        self.set_range(range)

    def set_feature(self, feature):
        '''It sets the feature'''
        self.__check_range(feature, self.__range)
        self.__feature = feature

    def get_feature(self):
        '''It returns the feature'''
        return self.__feature

    feature = property(get_feature, set_feature)

    def set_range(self, range):
        '''It sets the range'''
        self.__check_range(self.__feature, range)
        self.__range = range

    def get_range(self):
        '''It returns the range'''
        return self.__range

    range = property(get_range, set_range)

    def __is_sequence(self, feature):
        '''It returns True if the feature is a sequence type.
        
        A sequence type might be a list, a Seq or a RichSeq.
        A non-sequence type might be an str or an unicode
        '''
        if feature is None:
            return False
        if isinstance(feature, str) or isinstance(feature, unicode):
            return False
        try:
            len(feature)
        except TypeError:
            return False
        return True

    def __check_range(self, feature, range):
        #if the feature is an str we don't mind the legnth, because the
        #str is a label for all the range and it shouldn't be mapped to
        #every residue in the range.
        if not self.__is_sequence(feature):
            return
        if range is None:
            return
        if not len(feature) == len(range):
            raise RichFeatureError('Feature and Range lengths do not match.')

    def __getitem__(self, index):
        '''It returns a RichFeature that covers that slice.'''
        new_feat = RichFeature(id=self.id, type=self.type)
        #if there is no range associated to this RichFeature, we just return
        #the same feature
        range = self.range
        feature = self.feature
        if range is None:
            new_feat.feature = feature
            new_feat.range = None
            return new_feat

        #if there is a range
        if isinstance(index, int):
            requested_start = index
            requested_end = index
        else:
            if isinstance(index, slice):
                if index.start is None:
                    requested_start = self.range.start
                else:
                    requested_start = index.start
                if index.stop is None:
                    requested_end = self.range.end
                else:
                    requested_end = index.stop - 1

        #if the asked slice does not overlaps with the feature's range
        #we return None
        if not range.overlaps(BioRange(requested_start, requested_end +1)):
            return None

        #feature range               --------------
        #requested                            ---------
        #asked to the feature slice           -----
        #if the requested start or end is outside the range covered
        #by the feature we have to request only the part covered.
        if range.start >= requested_start:
            asked_start = range.start - requested_start
        else:
            asked_start = 0
        if range.end <= requested_end:
            asked_end = range.end -requested_start
        else:
            asked_end = requested_end - requested_start
        new_feat.range = BioRange(asked_start, asked_end)
        if not self.__is_sequence(feature):
            #if the feature is not a sequence type we just return it
            new_feat.feature = feature
        else:
            new_feat.feature = feature[asked_start:asked_end+1]

        return new_feat

    def __str__(self):
        toprint = 'id: ' + str(self.id) + '\n'
        toprint += 'type: ' + str(self.type) + '\n'
        toprint += 'range: ' + str(self.range) + '\n'
        toprint += 'feature: ' + str(self.feature) + '\n'
        return toprint

    def reverse(self, length=None):
        '''It reverses the feature.
        
        To do a proper reverse we need to know the length of the sequence in 
        which this RichFeature is. If None is provided we asume that the length
        is equal to the end of the feature.

        Keyword arguments:
        length -- the length of the sequence that holds this RichFeature,
                  an int (default the end of this RichFeature).
        '''
        range = self.range
        feature = self.feature
        #we reverse the range
        if range is not None:
            range.reverse(length)
        #if the feature is a sequence type we reverse it
        if self.__is_sequence(self.feature):
            try:
                #mutable objects
                feature.reverse()
            except AttributeError:
                #non mutable objects
                self.feature = feature[::-1]

    def complement(self):
        '''It complements the feature.'''
        feat = self.feature
        try:
            new_feat = feat.complement()
            #if the feature is not-mutable
            if type(new_feat) == type(feat):
                self.feature = new_feat
        except AttributeError:
            pass

import unittest

class RichFeatureTests(unittest.TestCase):
    def test_basic_functionality(self):
        feat =RichFeature(id='first_exon',type='exon',feature='This is an exon')
        assert feat.id == 'first_exon'
        assert feat.type == 'exon'
        assert feat.feature == 'This is an exon'

        range = BioRange(10, 14)
        feat = RichFeature(feature=[1,2,3,4,5], range=range)

        range = BioRange(10, 14)
        feat = RichFeature(feature='just an annotation', range=range)
        assert feat.feature == 'just an annotation'

    def test_exceptions(self):
        self.failUnlessRaises(RichFeatureError, RichFeature, 'id', 'type',
                                                      [1,2,3,4], BioRange(1,2))

    def test_slices(self):
        #if we have a feature without a range associated every slice will
        #mean the same, the whole feature
        #this is meant for features that apply to the whole sequence, any part
        #of the sequence will share the same feature
        feat =RichFeature(id='first_exon',type='exon',feature='This is an exon')
        feat2 = feat[:]
        assert feat.id == 'first_exon'
        assert feat.type == 'exon'
        assert feat.feature == 'This is an exon'

        #if the feature is a non-sequence one and the slice covers the
        #range we obtain, again, the same feature
        range = BioRange(10, 14)
        feat = RichFeature(feature='just an annotation', range=range)
        feat2 = feat[8:17]
        assert feat2.feature == 'just an annotation'
        #we have to take into account the change in the coordinate system
        #range            *****
        #feat   01234567890123456
        #feat2          0123456
        assert feat2.range.start == 2
        assert feat2.range.end == 6
        #another try
        range = BioRange(10, 14)
        feat = RichFeature(feature='just an annotation', range=range)
        feat2 = feat[9:16]
        assert feat2.feature == 'just an annotation'
        #we have to take into account the change in the coordinate system
        #range            *****
        #feat   01234567890123456
        #feat2           012345
        assert feat2.range.start == 1
        assert feat2.range.end == 5
        #and another more
        range = BioRange(10, 14)
        feat = RichFeature(feature='just an annotation', range=range)
        feat2 = feat[12:16]
        assert feat2.feature == 'just an annotation'
        #we have to take into account the change in the coordinate system
        #range            *****
        #feat   01234567890123456
        #feat2              012
        assert feat2.range.start == 0
        assert feat2.range.end == 2

        #if the feature is a non-sequence one and the slice does not cover the
        #range we obtain an empty feature
        range = BioRange(10, 14)
        feat = RichFeature(feature='just an annotation', range=range)
        feat2 = feat[9:14]
        assert feat2.feature == 'just an annotation'
        #we have to take into account the change in the coordinate system
        #range            *****
        #feat   0123456789012345
        #feat2           01234
        assert feat2.range.start == 1
        assert feat2.range.end == 4

        #the feature is a sequence type
        #range       01234
        #feat   01234567890
        #feat2        012
        range = BioRange(5,9)
        feat = RichFeature(feature=[0,1,2,3,4], range=range)
        feat2 = feat[6:9]
        assert feat2.range.start == 0
        assert feat2.range.end == 2
        assert len(feat2.feature) == 3
        assert feat2.feature[0] == 0
        assert feat2.feature[2] == 2

        #if we're asking for a slice that is not covered by the feature's range
        #we obtain an empty feature
        #the feature is a sequence type
        #range       01234
        #feat   0123456789012345
        #feat2              ----
        range = BioRange(5,9)
        feat = RichFeature(feature=[0,1,2,3,4], range=range)
        feat2 = feat[12:15]
        assert feat2 is None

    def test_ismutable(self):
        range = BioRange(5,9)
        feat = RichFeature(feature=[0,1,2,3,4], range=range)
        #              0123456789
        #feat               01234
        #feat.reverse()43210
        feat.reverse()
        assert feat.range.start == 0
        assert feat.range.end == 4
        assert len(feat.feature) == 5
        assert feat.feature[0] == 4
        assert feat.feature[4] == 0

        #this RichFeature might be inside a longer sequence
        #in this case the reverse is slighty different
        #seq           01234567890
        #feat               01234
        #feat.reverse() 43210
        range = BioRange(5,9)
        feat = RichFeature(feature=[0,1,2,3,4], range=range)
        feat.reverse(length=11)
        assert feat.range.start == 1
        assert feat.range.end == 5
        assert len(feat.feature) == 5
        assert feat.feature[0] == 4
        assert feat.feature[4] == 0

        #test with non mutable like Seq
        from Bio.Seq import Seq
        range = BioRange(6,9)
        feat = RichFeature(feature=Seq('ACGT'), range=range)
        feat.reverse(length=11)
        assert feat.range.start == 1
        assert feat.range.end == 4
        assert len(feat.feature) == 4
        assert feat.feature[0] == 'T'
        assert feat.feature[3] == 'A'

        #test with a  mutable Seq
        from Bio.Seq import MutableSeq
        range = BioRange(6,9)
        feat = RichFeature(feature=MutableSeq('ACGT'), range=range)
        feat.reverse(length=11)
        assert feat.range.start == 1
        assert feat.range.end == 4
        assert len(feat.feature) == 4
        assert feat.feature[0] == 'T'
        assert feat.feature[3] == 'A'

        #test with nonsequence
        from Bio.Seq import Seq
        range = BioRange(6,9)
        feat = RichFeature(feature='an annot', range=range)
        feat.reverse(length=11)
        assert feat.range.start == 1
        assert feat.range.end == 4
        assert feat.feature == 'an annot'

        #complement
        #this feature can't be complemented, it should remain the same
        range = BioRange(5,9)
        feat = RichFeature(feature=[0,1,2,3,4], range=range)
        feat.complement()
        assert feat.range.start == 5
        assert feat.range.end == 9
        assert len(feat.feature) == 5
        assert feat.feature[0] == 0
        assert feat.feature[4] == 4

        #test with non mutable like Seq
        from Bio.Seq import Seq
        range = BioRange(6,9)
        feat = RichFeature(feature=Seq('ATGC'), range=range)
        feat.complement()
        assert feat.range.start == 6
        assert feat.range.end == 9
        assert len(feat.feature) == 4
        assert feat.feature[0] == 'T'
        assert feat.feature[3] == 'G'

        #test with a mutableSeq
        from Bio.Seq import MutableSeq
        range = BioRange(6,9)
        feat = RichFeature(feature=MutableSeq('ATGC'), range=range)
        feat.complement()
        assert feat.range.start == 6
        assert feat.range.end == 9
        assert len(feat.feature) == 4
        assert feat.feature[0] == 'T'
        assert feat.feature[3] == 'G'

def main():
    unittest.main()

if __name__ == '__main__':
    main()
