#!/usr/bin/python
import Bio.Seq
from RichFeature import RichFeature

class RichSeqError(Exception):
    def __init__(self, value):
        self.parameter = value
    def __str__(self):
        return repr(self.parameter)

class RichSeq(object):
    '''RichSeq represents a sequence and its annotations.

    A sequence can have, besides a sequence, a quality, an id, a name, an 
    a list of features (RichFeature).
    RichSeq is meant to be a drop in replace for SeqRecord, but there are
    two main differences between RichSeq and SeqRecord:
        - RichSeq supports quality
        - RichSeq has a __getitem__ method equivalent to the one present in
        Seq
    A subsequence from RichSeq can be obtained using the syntax some_seq[10:20].
    '''
    def __init__(self, id=None, seq=None, qual=None, name=None,
                 description=None, features=None):
        '''The RichSeq constructor.

        Keyword arguments:
        id -- the sequence id (default None)
        seq -- the sequence, a Seq object (default None)
        qual -- an int list (default [])
        name -- the sequence name (default None)
        description -- the sequence description (default None)
        features -- the sequence features, a list of RichFeature (default [])
        '''
        self.id = id
        self.__seq = None
        self.__qual = None
        self.set_seq(seq)
        self.set_qual(qual)
        self.name = name
        self.description = description
        self.set_features(features)

    def __len__(self):
        '''It returns the length of the sequence.

        If there is a seq or a qual it returns the length, otherwise it returns
        0.
        '''
        if self.seq is not None:
            return len(self.seq)
        if self.qual is not None:
            return len(self.qual)
        #if there are feature we look for the largest one
        end = 0
        for feat in self.features:
            range = feat.range
            if feat.range is not None:
                if end < range.end + 1:
                    end = range.end + 1
        return end

    def set_seq(self, seq):
        '''It sets the sequence.

        If there is a quality, it checks that the seq len is equal to the qual
        len.

        Keyword arguments:
        seq -- the sequence, an str or a Seq object
        '''
        if isinstance(seq, str) or isinstance(seq, unicode):
            seq = Bio.Seq.Seq(seq)
        self.__check_seq_qual_length(seq, self.__qual)
        self.__seq = seq

    def get_seq(self):
        '''It returns the sequence.'''
        return self.__seq

    seq = property(get_seq, set_seq)

    def set_qual(self, qual):
        '''It sets the quality.
        
        The quality can be set even if there are no sequence.
        It checks that the quality and seq length match.
        '''
        #the quality should be a list
        if qual is not None and not isinstance(qual, list):
            raise RichSeqError('The quality should be a list')
        self.__check_seq_qual_length(self.__seq, qual)
        self.__qual = qual

    def get_qual(self):
        '''It returns the quality.'''
        return self.__qual

    def __check_seq_qual_length(self, seq, qual):
        if seq is None:
            return
        if qual is None:
            return
        if len(seq) <> len(qual):
            raise RichSeqError('Sequence and Quality lengths do not match.')

    qual = property(get_qual, set_qual)

    def set_features(self, features):
        '''It sets the features.
        
        It check that the features is a list
        '''
        if features is None:
            features = []
        if not isinstance(features, list):
            raise RichSeqError('Features should be a list')
        #if there is a seq or a qual the features can not be larger than that
        if self.seq is not None or self.qual is not None:
            l = len(self)
            for feat in features:
                range = feat.range
                if range is not None and range.end + 1 > l:
                    raise RichSeqError('A feature is outside the range of the sequence.')
        self.__features = features

    def get_features(self):
        '''It returns the features'''
        return self.__features

    features = property(get_features, set_features)

    def __getitem__(self, index):
        '''It returns an item or slice as a RichSeq'''
        #first the non-sequence properties
        new_seq = RichSeq(id=self.id, name=self.name,
                                                description=self.description)
        #now the sequence properties
        #sequence
        if self.__seq is None:
            seq = None
        else:
            seq = self.__seq[index]
        new_seq.seq = seq
        #quality
        if self.__qual is None:
            qual = None
        else:
            qual = self.__qual[index]
        new_seq.qual = qual
        #features
        feats = []
        for feat in self.__features:
            sub_feat = feat[index]
            if sub_feat is not None:
                feats.append(sub_feat)
        new_seq.features = feats
        return new_seq

    def reverse(self):
        '''It reverses the sequence.'''
        #reverse the sequence
        if self.seq is not None:
            self.seq = self.seq[::-1]
        #the quality
        if self.qual is not None:
            self.qual = self.qual[::-1]
        #the features
        l = len(self)
        for feat in self.features:
            feat.reverse(l)

    def complement(self):
        '''It complements the sequence'''
        #complement the sequence
        seq = self.seq
        if seq is not None:
            #the sequence can be mutable or not mutable
            new_seq = seq.complement()
            if type(new_seq) == type(seq):  #if it was not mutable
                self.seq = new_seq
        #the features
        for feat in self.features:
            feat.complement()

    def reverse_complement(self):
        '''It reverses and complements the sequence'''
        self.reverse()
        self.complement()

import unittest
from BioRange import BioRange

class SequenceTests(unittest.TestCase):
    def test_basic_functionality(self):
        #default alphabet
        default_alpha = Bio.Seq.Seq('ACTG').alphabet
        #basic sequence creation
        #from a string, an alphabet should be added
        seq = RichSeq(seq='ACTG')
        assert seq.seq.tostring() == 'ACTG'
        assert seq.seq.alphabet == default_alpha
        assert seq.qual is None
        assert seq.id is None

        #seq init using a sequence string
        ss = Bio.Seq.Seq('ACTG') #the default alphabet is added
        seq = RichSeq(seq=ss)
        assert seq.seq.tostring() == 'ACTG'
        assert seq.seq.alphabet == default_alpha
        assert seq.qual is None
        assert seq.id is None

        from Bio.Alphabet import IUPAC
        #using an alphabet
        ss = Bio.Seq.Seq('ACTG', IUPAC.unambiguous_dna)
        seq = RichSeq(seq=ss)
        assert seq.seq.tostring() == 'ACTG'
        assert seq.seq.alphabet == IUPAC.unambiguous_dna
        assert seq.qual is None
        assert seq.id is None

        #with quality
        seq = RichSeq(seq='ACTG', qual=[60,60,60,60])
        assert seq.seq.tostring() == 'ACTG'
        assert seq.seq.alphabet == default_alpha 
        assert len(seq.qual) == 4
        assert seq.id is None

        #only with quality
        seq = RichSeq(qual=[60,60,60,60])
        assert seq.seq is None
        assert len(seq.qual) == 4
        assert seq.id is None

        #with an id
        seq = RichSeq(id='test_seq', seq='ACTG', qual=[60,60,60,60])
        assert seq.seq.tostring() == 'ACTG'
        assert seq.seq.alphabet == default_alpha
        assert len(seq.qual) == 4
        assert seq.id == 'test_seq'

        #other attributes
        seq = RichSeq(id='test_id', seq='ACTG', qual=[60,60,60,60],
                    name='test_name', description='just a test seq')
        assert seq.seq.tostring() == 'ACTG'
        assert seq.seq.alphabet == default_alpha
        assert len(seq.qual) == 4
        assert seq.id == 'test_id'
        assert seq.name == 'test_name'
        assert seq.description == 'just a test seq'

    def test_feature(self):
        #a feature can be just a string with an annotation
        feat = RichFeature(type='annotation', feature='just an annotation')
        seq = RichSeq(seq='ACTG', features=[feat])
        assert seq.seq.tostring() == 'ACTG'
        assert isinstance(seq.features[0], RichFeature) == True
        assert seq.features[0].feature == 'just an annotation'
        assert seq.features[0].type == 'annotation'
        assert seq.features[0].range is None

        #a feature can have a range associated
        range = BioRange(start=1,end=2)
        feat = RichFeature(type='annotation', range=range,
                            feature='an annotation in a range, e.g. a domain')
        seq = RichSeq(seq='ACTG', features=[feat])
        assert seq.features[0].range.start == 1

        #a ranged feature can be also a Seq or a RichSeq or any other object
        #that maps a property for every position.
        #this converts RichSequence into a recursive object
        #In that way a subsequence can be asssociated to a fragment of the 
        #sequence. It can be used, for instance to add the sequence of an
        #exon.
        range = BioRange(start=1,end=2)
        feat = RichFeature(type='subsequence', feature=RichSeq(seq='CT'),
                                                                    range=range)
        seq = RichSeq(seq='ACTG', features=[feat])
        assert isinstance(seq.features[0], RichFeature) == True
        assert seq.features[0].feature.seq.tostring() == 'CT'
        assert seq.features[0].range.start == 1

    def test_slice(self):
        #default alphabet
        default_alpha = Bio.Seq.Seq('ACTG').alphabet
        #this is a key point, the slice should be applied to RichSeq, not
        #to seq, in that way the quality also gets sliced
        seq = RichSeq(id='test_seq', seq='ACTG', qual=[60,60,60,60])
        seq2 = seq[0:2]
        assert isinstance(seq2, RichSeq)
        assert seq2.seq.tostring() == 'AC'
        assert seq2.seq.alphabet == default_alpha
        assert len(seq2.qual) == 2
        #i'm not sure about the following, maybe it should be test_seq_0-1,
        #but in that case, what about the dbxrefs?
        assert seq.id == 'test_seq'
        
        #with some extra attributes
        seq = RichSeq(id='test_id', seq='ACTG', qual=[60,60,60,60],
                    name='test_name', description='just a test seq')
        assert seq.seq.tostring() == 'ACTG'
        assert seq.seq.alphabet == default_alpha
        assert len(seq.qual) == 4
        assert seq.id == 'test_id'
        assert seq.name == 'test_name'
        assert seq.description == 'just a test seq'

    def test_slice_with_ranged_features(self):
        #a feature with a range associated
        range = BioRange(start=1,end=2)
        annot ='an annotation in a range, e.g. a domain'
        feat = RichFeature(type='annotation', range=range, feature=annot)
        seq = RichSeq(seq='ACTGACTG', features=[feat])
        #the subsequence covers the range of the feature
        seq2 = seq[1:3]
        assert seq2.features[0].range.start == 0
        assert seq2.features[0].range.end == 1
        assert seq2.features[0].feature == annot

        #the subsequence does not cover the feature
        seq2 = seq[3:4]
        assert not seq2.features   #empty list

        #the subsequence covers only partialy the feature
        #we don't know how to split the subsequence, so we leave it as it is.
        #if we want a different behaviour we should create a new feature
        #type derived from str, but with a new __getitem__ able to create 
        #somethig like 'Exon 1 fragment from 10 to 20'
        seq2 = seq[1:2]
        assert seq2.features[0].range.start == 0
        assert seq2.features[0].range.end == 0
        assert seq2.features[0].feature == annot

        #a ranged feature can be also a Seq or a RichSeq or any other object
        #that maps a property for every position.
        #this converts RichSequence into a recursive object
        #In that way a subsequence can be asssociated to a fragment of the 
        #sequence. It can be used, for instance to add the sequence of an
        #exon.
        range = BioRange(start=1,end=2)
        feat = RichFeature(type='subsequence', feature=RichSeq(seq='CT'),
                                                                    range=range)
        seq = RichSeq(seq='ACTG', features=[feat])
        #covers the subsequence
        seq2 = seq[1:3]
        assert isinstance(seq.features[0], RichFeature)
        assert seq2.features[0].feature.seq.tostring() == 'CT'
        assert seq2.features[0].range.start == 0
        assert seq2.features[0].range.end == 1
        #covers only partialy
        seq3 = seq[1:2]
        assert isinstance(seq3.features[0], RichFeature)
        assert seq3.features[0].feature.seq.tostring() == 'C'
        assert seq3.features[0].range.start == 0
        assert seq3.features[0].range.end == 0

    def test_ismutable(self):
        #Rich sequence is mutable an it can be reversed and complemented
        #reverse
        default_alpha = Bio.Seq.Seq('ACTG').alphabet
        annot ='an annotation in a range, e.g. a domain'
        range = BioRange(2,3)
        feat = RichFeature(type='annotation', range=range, feature=annot)
        seq = RichSeq(id='test_id', seq='ACTG', qual=[60,60,40,40],
                    name='test_name', description='just a test seq',
                    features=[feat])
        seq.reverse()
        assert seq.seq.tostring() == 'GTCA'
        assert seq.seq.alphabet == default_alpha
        assert len(seq.qual) == 4
        assert seq.qual[0] == 40
        assert seq.id == 'test_id'
        assert seq.name == 'test_name'
        assert seq.description == 'just a test seq'
        new_feat = seq.features[0]
        assert new_feat.feature == annot
        assert new_feat.range.start == 0
        assert new_feat.range.end == 1

        #reverse with a mutable seq
        from Bio.Seq import MutableSeq
        seq = RichSeq(id='test_id', seq=MutableSeq('ACTG'), qual=[60,60,40,40])
        seq.reverse()
        assert seq.seq.tostring() == 'GTCA'

        #complement
        range = BioRange(2,3)
        feat = RichFeature(type='annotation', range=range, feature=annot)
        seq = RichSeq(id='test_id', seq='ACTG', qual=[60,60,40,40],
                    name='test_name', description='just a test seq',
                    features=[feat])
        seq.complement()
        assert seq.seq.tostring() == 'TGAC'
        assert seq.seq.alphabet == default_alpha
        assert len(seq.qual) == 4
        assert seq.qual[0] == 60
        assert seq.id == 'test_id'
        assert seq.name == 'test_name'
        assert seq.description == 'just a test seq'
        new_feat = seq.features[0]
        assert new_feat.feature == annot
        assert new_feat.range.start == 2
        assert new_feat.range.end == 3
        #complent with a mutable seq
        from Bio.Seq import MutableSeq
        seq = RichSeq(id='test_id', seq=MutableSeq('ACTG'), qual=[60,60,40,40])
        seq.complement()
        assert seq.seq.tostring() == 'TGAC'

        #reverse and complement
        seq = RichSeq(id='test_id', seq='ACTG', qual=[60,60,40,40],
                    name='test_name', description='just a test seq')
        seq.reverse_complement()
        assert seq.seq.tostring() == 'CAGT'
        assert seq.seq.alphabet == default_alpha
        assert len(seq.qual) == 4
        assert seq.qual[0] == 40
        assert seq.id == 'test_id'
        assert seq.name == 'test_name'
        assert seq.description == 'just a test seq'

    def test_copy(self):
        seq = RichSeq(id='test_id', seq='ACTG')
        seq2 = seq[:]
        assert seq is not seq2
        #it should be a deepcopy of every subfeature and annotation
        #otherwise modifying one sequence would affect the other.
        seq2.id = 'test_id2'
        assert seq.id == 'test_id'

    def test_length(self):
        seq = RichSeq(id='test_id', seq='ACTG')
        assert len(seq) == 4
        seq = RichSeq(id='test_id', qual=[0,1,2,3])
        assert len(seq) == 4
        range = BioRange(2,3)
        feat = RichFeature(type='annotation', range=range, feature='ORF')
        seq = RichSeq(id='test_id', features=[feat])
        assert len(seq) == 4

    def test_errors(self):
        #quality and seq length do not match
        self.failUnlessRaises(RichSeqError, RichSeq, seq='ACTG', qual=[10])
        #quality is not a list
        self.failUnlessRaises(RichSeqError, RichSeq, seq='A', qual=10)
        range = BioRange(2,13)
        feat = RichFeature(type='annotation', range=range, feature='ORF')
        #features outside the range covered by the sequence
        self.failUnlessRaises(RichSeqError, RichSeq, seq='ACTG',features=[feat])
        #feature is not a list
        self.failUnlessRaises(RichSeqError, RichSeq, features=feat)
        
    def test_add(self):
        #TODO seq1 + seq2
        #el alineamiento debe heredar de richseq. anyade metodos para append
        #sequence modifica el getitem
        #permite recuperar secuencia por id
        pass


def main():
    unittest.main()


if __name__ == '__main__':
    main()
